from sympy import *
from functools import reduce
from typing import Union

def _extract_vars(expr):
    return [e[1] for e in sorted(list(map(lambda x: (str(x),x), expr.free_symbols)))]


def taylor(expr, vars, degree):
    if len(vars) != 2 and len(vars) != 4:
        raise Exception("taylor is only defined for one or two variables")
    if len(vars) == 2:
        return series(expr, vars[0], vars[1], n=degree).removeO()
    P_n = S(0)
    x, x0, y, y0 = vars
    for i in range(degree):
        for j in range(degree - i):
            P_n += (
                expr.diff(x, i).diff(y, j).subs([(x, x0), (y, y0)])
                * (x - x0) ** i
                * (y - y0) ** j
                / (factorial(i) * factorial(j))
            )

    return P_n


def gradient(expr, var=None):
    var = _extract_vars(expr) if var is None else var
    if len(var) == 1:
        return expr.diff(var[0])
    return Matrix([expr.diff(v) for v in var])


def hessian(expr, var=None):
    var = _extract_vars(expr) if var is None else var
    return Matrix(
        [[reduce(lambda e, s: e.diff(s), [x, y], expr) for x in var] for y in var]
    )


def _extract_field_vars(V, var, namefunc):
    var = _extract_vars(V) if var is None else var
    
    # if 3 variables cannot be found, dummy variable is returned,
    # such that differentiating wrt. to it results in 0.
    while len(var) < 3:
        var.append(symbols("_dummy"))
        
    return var


def div(V, var=None):
    var = _extract_field_vars(V, var, "div")

    return sum([V[i].diff(f) for i, f in enumerate(var)])


def rot(V, var=None):
    var = _extract_field_vars(V, var, "rot")

    return Matrix(
        [
            V[2].diff(var[1]) - V[1].diff(var[2]),
            V[0].diff(var[2]) - V[2].diff(var[0]),
            V[1].diff(var[0]) - V[0].diff(var[1]),
        ]
    )


def dsolve(ODE: Union[Eq, list, Matrix], ics=None) -> dict:
    """ A wrapper for the sympy dsolve-function. Instead of a list of equations
        denoting the solution, this function returns a dictionary of solutions.

        // This makes it easier to substitute solution in the places it's needed

        ## Input:
            ODE: one or multiple differential equations to be solved by sympy

        ## Returns:
            sol: a dictionary containing solutions to 'ODE'
        
    """
    
    # Common problem for people is making an single equation, but with both sides
    # being Matrices. This alleviates that issue, and lets users think less
    if type(ODE) == Eq and type(ODE.lhs) == type(ODE.rhs) == Matrix:
            sol = dsolve(ODE.lhs - ODE.rhs, ics=ics)
    else:
        sol = dsolve(ODE, ics=ics)

    return {eq.lhs : eq.rhs for eq in sol}


def l2_norm( v: Matrix ):
    """ Computes the l2-norm for a Matrix-class object, without using
        absolute values on entries, for easier simplification and integration

        ## Input:
            v: A sympy Matrix
    """
    return sqrt(sum(x**2 for x in v))


from sympy import latex, Matrix
from sympy.external import import_module
from spb.backends.base_backend import Plot as BP
from spb.defaults import TWO_D_B, THREE_D_B, cfg
from spb.functions import _set_labels
from spb.series import BaseSeries
from spb.utils import (
    _plot_sympify,
    _unpack_args_extended,
    _split_vector,
    _is_range,
    _instantiate_backend,
)


class ArrowSeries(BaseSeries):
    """Represent a vector field."""

    is_vector = True
    is_slice = False
    is_streamlines = False
    _allowed_keys = []

    def __init__(self, start, direction, label=None, **kwargs):

        self.start = start
        self.direction = direction

        self._label = f"{start}->{direction}" if label is None else label
        self._latex_label = latex(f"{start}->{direction}") if label is None else label

        self.expr = kwargs.get("expr", "")
        self.use_cm = kwargs.get("use_cm", False)
        self.color_func = kwargs.get("color_func", None)
        # NOTE: normalization is achieved at the backend side: this allows to
        # obtain same length arrows, but colored with the actual magnitude.
        # If normalization is applied on the series get_data(), the coloring
        # by magnitude would not be applicable at the backend.
        self.normalize = kwargs.get("normalize", False)

        self.rendering_kw = kwargs.get("quiver_kw", kwargs.get("rendering_kw", dict()))

        self._set_use_quiver_solid_color(**kwargs)
        self._init_transforms(**kwargs)

    def _set_use_quiver_solid_color(self, **kwargs):
        # NOTE: this attribute will inform the backend wheter to use a
        # color map or a solid color for the quivers. It is placed here
        # because it simplifies the backend logic when dealing with
        # plot sums.
        self.use_quiver_solid_color = (
            True
            if ("scalar" not in kwargs.keys())
            else (
                False if (not kwargs["scalar"]) or (kwargs["scalar"] is None) else True
            )
        )

    def get_data(self):

        return *self.start, *self.direction


def quiver(*args, **kwargs):

    np = import_module("numpy")

    # format if numbers are entered directly instead of lists
    num_single = 0
    args = list(args)
    for i in range(len(args)):
        if type(args[i]) in [int, float]:
            num_single += 1
        else:
            try:
                args[i] = float(args[i])
                num_single += 1
            except:
                break
    if num_single == 4:
        if len(args) > 4:
            args = [[args[0], args[1]], [args[2], args[3]], *args[4:]]
        else:
            args = [[args[0], args[1]], [args[2], args[3]]]
    elif num_single == 6:
        if len(args) > 6:
            args = [[args[0], args[1], args[2]], [args[3], args[4], args[5]], *args[6:]]
        else:
            args = [[args[0], args[1], args[2]], [args[3], args[4], args[5]]]
    elif num_single != 0:
        raise f"Error! Wrong format used in quiver. Got {num_single} arguments \
            that could be start or direction vector coordinates!"

    # split arguments into vectors and other arguments
    point_args = []
    otherargs = []
    for i in range(len(args)):
        if type(args[i]) in [list, np.ndarray, tuple, type(Matrix())]:
            newpoint = np.array(args[i]).flatten()
            point_args.append(newpoint)
        elif type(args[i]) == dict:
            kwargs.setdefault("rendering_kw", args[i])
        else:
            otherargs.append(args[i])
    assert (
        len(point_args) == 2
    ), f"Error! Start and direction vectors must be provided (only two vectors)!"

    try:
        point_args = np.array(point_args, dtype=float)
        assert point_args.shape[-1] in [
            2,
            3,
        ], "Error! Start and direction vectors must be 2D or 3D!"
        args = otherargs
    except:
        raise f"Error! Wrong format used in quiver. \
        Got {point_args[0]} as starting point(s) and {point_args[1]} as ending point(s)!"

    # want structure to be [list of starts, list of ends]
    # where list of starts could be [start1, start2, ...], either 2D or 3D points
    if len(point_args.shape) == 2:
        point_args = point_args[:, None, :]

    labels = kwargs.pop("label", [])
    if type(labels) == str:
        labels = [labels]
    assert type(labels) == list, f"Error! Label must be a list or a string!"
    assert len(labels) in [
        0,
        point_args.shape[1],
    ], f"Error! Number of labels must be equal to number of arrows, or empty list!"
    if labels == []:
        labels = [None] * point_args.shape[1]

    kwargs.setdefault("legend", True)
    params = kwargs.get("params", None)
    is_interactive = False if params is None else True
    kwargs["is_interactive"] = is_interactive
    if is_interactive:
        raise NotImplementedError("Interactive quiver plots are not yet implemented!")
        from spb.interactive import iplot

        kwargs["is_vector"] = True
        return iplot(*args, **kwargs)

    series = [
        ArrowSeries(start, stop, *otherargs, label=label, **kwargs)
        for start, stop, label in zip(point_args[0, :, :], point_args[1, :, :], labels)
    ]

    rendering_kw = kwargs.pop("rendering_kw", None)

    # if 2D
    if point_args.shape[-1] == 2:
        Backend = kwargs.pop("backend", TWO_D_B)
        for i in range(len(series)):
            series[i].rendering_kw["angles"] = "xy"
            series[i].rendering_kw["scale_units"] = "xy"
            series[i].rendering_kw["scale"] = 1
            series[i].is_2Dvector = True
    else:
        Backend = kwargs.pop("backend", THREE_D_B)
        for i in range(len(series)):
            series[i].is_3Dvector = True

    _set_labels(series, labels, rendering_kw)

    # if multiple arrows are given, plot them all in one plot
    show = kwargs.pop("show", True)
    B = _instantiate_backend(Backend, series[0], show=False, **kwargs)
    for i in range(1, len(series)):
        B.extend(_instantiate_backend(Backend, series[i], show=False, **kwargs))
    if show:
        B.show()
    return B



from sympy import Matrix
from sympy.external import import_module
from spb.functions import plot_list, plot3d_list


def scatter(*args, **kwargs):

    np = import_module("numpy")
    warnings = import_module("warnings")

    def warning_formatter(msg, *args, line=None, **kwargs):
        return str(msg)

    warnings.formatwarning = warning_formatter

    # format all arguments
    dim = 0
    args = list(args)
    otherargs = []
    matrixlist = []
    for i in range(len(args)):
        if type(args[i]) in (list, tuple):
            args[i] = np.array(args[i])
            dim += 1
        elif type(args[i]) == np.ndarray:
            if len(args[i].shape) > 1:
                warnings.warn(
                    f"np.array with more than 1 dimension ({len(args[i].shape)} dimensions was found) has been flattened!"
                )
            args[i] = args[i].flatten()
            dim += 1
        elif type(args[i]) in [float, int]:
            args[i] = np.array([args[i]])
            dim += 1
        elif type(args[i]) == type(Matrix()):
            matrixlist.append(args[i])
        else:
            try:
                # symbolic expression not multiplied fully
                args[i] = np.array([float(args[i])])
                dim += 1
            except:
                otherargs.append(args[i])

    # if entries are matricies, get them into right format
    if len(matrixlist) > 0:
        assert dim == 0, "Cannot mix matrix and non-matrix arguments!"
        firstshape = matrixlist[0].shape
        assert len(firstshape) in [1, 2], "Matrix must be 1D or 2D!"
        assert (
            min(firstshape) == 1
        ), "Matrix must not have multiple dimensions different from size 1!"
        dim = max(firstshape)
        newargs = [np.array([])] * dim
        for i in range(len(matrixlist)):
            assert (
                matrixlist[i].shape == firstshape
            ), "All matrices must have the same shape!"
            for j in range(len(matrixlist[i])):
                newargs[j] = np.append(newargs[j], matrixlist[i][j])
        args = newargs
        args.extend(otherargs)

    assert dim in [2, 3], "scatterplot only supports 2D and 3D plots"
    if dim == 2:
        rendering_kw = kwargs.pop("rendering_kw", {})
        markersize = rendering_kw.pop("s", None)
        if markersize is not None:
            rendering_kw.setdefault("markersize", markersize)
        return plot_list(*args, is_point=True, rendering_kw=rendering_kw, **kwargs)
    else:
        return plot3d_list(*args, is_point=True, **kwargs)
